
"""High level system administration functions.

All of these are "shell" scripts and may throw RunFail or other exceptions.

Import this after arg parsing is done (for global state).
"""
from shell import *

verbose, testmode = verbose_testmode()

def partition_format_and_mount_drive(target_dev_name, mount_root_at, drive_name=None):
    """This is currently hard-coded for my preferred partitioning scheme for a UEFI main/boot disk
    but could be further parameterized.

    target_dev_name is the device to format, as in /dev/sda
    mount_root_at   is the root dir to mount that on, as in /mnt/restore
    drive_name      is a word, as in 'root_backup', to write to the file <mount_root_at>/drive.name, or None to skip.
    """

    #================================================================
    # Sanity checks...
    #
    echo("Preparing to partition and format %s"%(target_dev_name,))

    if not isblockspecial(target_dev_name):
        bail("ERROR: %s is not a block special device."%(target_dev_name,))

    target_dev = find_device_named(target_dev_name)

    if target_dev is None:
        bail("ERROR: can't find %s in lsblk output."%(target_dev_name,))
 
    if device_parent(target_dev):
        bail("ERROR: %s is not a top-level device."%(target_dev_name,))

    for part in device_partitions(target_dev):
        if part.get('mountpoint'):
            bail("ERROR: %s is currently mounted (at %s)!"%(device_name(part),part.get('mountpoint')), test_ignore=True)

    sizeGiB = int(target_dev.get('size', 0))/(1024.**3)

    echo("The size of %s is %.2f GiB"%(target_dev_name, sizeGiB))

    if sizeGiB < 30.:
        bail("ERROR: Probably too small!  Sure that's the right device?")

    if not user_wants_to("destroy everything on %s, re-partition and format it"%(target_dev_name,)):
        bail("Aborting...", 0)

    #
    # End of sanity checks.
    #================================================================
    # Partition the drive...
    #
    run(["parted", "--align=optimal", target_dev_name,
             "mklabel", "gpt",
             "mkpart",     "ESP", "fat32",     "0%", "128MiB",
             "mkpart", "primary",  "ext4", "128MiB",  "32GiB",
             "mkpart", "primary",           "32GiB",   "100%",
             "set", "1", "boot", "on"])
    run(['partprobe', target_dev_name])
    sleep(1)
    if testmode:
        run(['lsblk'], safe=True)   # Can't run the other unless root, so this lets us test more safely.
    else:
        run(['parted', target_dev_name, 'unit', 'MiB', 'print'], safe=True)

    get_devices(True)   # Force re-load
    target_dev = find_device_named(target_dev_name)
    partitions = device_partitions(target_dev)

    if testmode:
        if len(partitions) < 2:
            bail("ERROR: %r should have at least two partitions but doesn't.  See lsblk?"%(target_dev_name,))
    else:
        if len(partitions) != 3:
            bail("ERROR: %r should have three partitions but doesn't.  See lsblk?"%(target_dev_name,))

    partnames = [device_name(part) for part in partitions]
    assert '/dev/' not in partnames

    echo("Partition names are: %r"%(partnames,))

    if not user_wants_to("continue (will format the first two partitions)"):
        bail("Aborting...")

    run(["mkfs.fat", "-F32", partnames[0]]        )
    run(["mkfs.ext4",        partnames[1]]        )
    run(['mount',            partnames[1], reroot(mount_root_at, '/')])
    run(['mkdir', '-p',                    reroot(mount_root_at, '/boot/efi')])
    run(['mount',            partnames[0], reroot(mount_root_at, '/boot/efi')])
    sleep(1)

    get_devices(True)   # Force re-load

    #
    # Create /drive.name for the root partition:
    #
    if drive_name:
        target_dname_path = reroot(mount_root_at, '/drive.name')
        if testmode:
            echo("TEST MODE: would write %r to %s"%(drive_name, target_dname_path))
        else:
            echo("Creating %s"%(target_dname_path,))
            with open(target_dname_path, 'w') as fl:
                fl.write(drive_name + "\n")

def fix_fstab_UUIDs(root_dir):
    """This looks for any partitions currently mounted under root_dir (including root_dir itself)
        and updates their UUIDs in root_dir/etc/fstab to point to the correct partition.

    I.e., if /mnt/restore is where a restored system root currently lives, this will update
        /mnt/restore/etc/fstab so that the partition currently mounted on /mnt/resstore is
        mapped to mount on '/', and so on (provided such entries already exist in fstab).

    The fstab is otherwise unmolested, and all formatting and flags are preserved even for the
        updated lines.
    """
    fstab_path = reroot(root_dir, "/etc/fstab")

    if not user_wants_to("update %s with new UUIDs"%(fstab_path,)):
        bail("Aborting...")

    fstab = file_body(fstab_path).split('\n')

    for dev in find_devices('mountpoint', root_dir, match_func=path_startswith):

        t_mountpoint = dev['mountpoint']
        mountpoint   = deroot(root_dir, t_mountpoint)

        uuid = dev.get('uuid')
        if not uuid:
            bail("Unable to find UUID for %s (%s)"%(mountpoint, t_mountpoint,))
        uuid = "UUID="+uuid # fstab format

        new_fstab = []
        for line in fstab:
            if line.strip() and not line.strip().startswith('#'):
                fields = line.strip().split()
                f_uuid       = fields[0]
                f_mountpoint = fields[1]
                if f_mountpoint == mountpoint:
                    echo("%12s: %s -> %s"%(mountpoint, f_uuid, uuid))
                    line = line.replace(f_uuid, uuid)
                elif f_uuid == uuid:
                    echo("WARNING: UUID for %s previously mapped to %s.  Commenting out."%(mountpoint, f_mountpoint))
                    line = "# " + line
            new_fstab.append(line)
        fstab = new_fstab

    fstab = '\n'.join(fstab)

    if testmode:
        echo("Would write to %s:"%(fstab_path,))
        print("---")
        print(fstab)
        print("---")
    else:
        echo("Saving %s"%(fstab_path,))
        with open(fstab_path, 'w') as fl:
            fl.write(fstab)

def install_boot_loader(root_path, nvram=False):
    """This chroots into root_path and runs grub-install and update-grub.

    If you want to set the local computer's nvram to use the new install, set nvram=True,
        otherwise the (UEFI) nvram will be untouched.

    Because it's awkward to recover from a chroot, this function either exit(0)'s or
        throws an exception.  In the latter case it's not known whether the chroot
        has already happened so in general you should exit as promptly as possible.
    """
    specials = ['/dev', '/dev/pts', '/sys', '/proc', '/run', '/tmp']

    try:
        mounted = []

        def cleanup():
            for tdir in reversed(mounted):
                try:
                    run(['umount', tdir], safe=True)    # Not super safe, but we're going to treat mount/umount as non-destructive here...
                except Exception as e:
                    echo("WARNING: Unable to umount %s (%s)"%(tdir, e))

        #
        # Mount all the special dirs under the backup root:
        #
        for sdir in specials:
            tdir = reroot(root_path, sdir)
            run(['mount', '--bind', sdir, tdir], safe=True)
            mounted.append(tdir)

        echo("CHROOT %s"%(root_path,))

        # Chroot to our new environ:
        os.chroot(root_path)

        # Tell cleanup we're in chrooted environ:
        mounted = specials

        os.chdir('/')    # If we don't call chdir, we're left outside of the reachable tree and bad things happen.

        if nvram:
            run(['grub-install'])
        else:
            run(['grub-install', '--no-nvram'])

        if testmode:
            run(['grub-mkconfig', '-o', '/tmp/bootbak.grub.cfg'], safe=True)
        else:
            run(['update-grub'])

    finally:
        cleanup()

    # If we've made it our boot disk and it was the 'restore' drive, we can call it 'root' now:
    if nvram and file_body_stripped('/drive.name', 'restore')=='restore':
        if testmode:
            echo("Would write 'root' to /drive.name")
        else:
            with open('/drive.name', 'w') as fl:
                fl.write("root\n")
            echo("Updated /drive.name to 'root'")

    echo("All done.")
    exit(0)

