#!/usr/bin/python3

import os, stat, subprocess, json

from sys        import argv, exit
from time       import sleep

#===================================================================
# Arg parsing (support function with example usage below):

args   = argv[1:]
run_as = argv[0].split('/')[-1]

_verbose  = 0
_testmode = 0
_argsparsed = False # Just for sanity checking

def nextarg():
    return args.pop(0)

def argparse(parseflag, parsearg, use):
    """Parse command-line arguments.

    parseflag(arg) will be passed args (one at a time) which start with '-' except those after a '--' (which declares no more flags)
    parsearg(arg) will be passed args that don't start with '-', or any after the first '--'.

    Either can be None if none such expected.
    Both should return True on error (unrecognized flag or unexpected argument) or None on success.
    Call nextarg() to fetch the next argument, for, e.g., flags that are followed by one or more parameters.

    Returns (verbose, testmode) since otherwise they would need to be re-imported.

    See below for example code.
    """
    global _verbose, _testmode, _argsparsed

    assert not _argsparsed

    moreflags = True

    while args:
        arg = nextarg()

        if moreflags and arg.startswith('-'):

            if arg == '-use' or arg == '-h':
                print(use)
                exit(0)
            elif arg == '-v':
                _verbose += 1
            elif arg == '-t':
                _testmode += 1
            elif arg == '--':
                moreflags = False
            elif parseflag is None or parseflag(arg):
                print("Unknown option %r"%(arg,))
                exit(1)

        elif parsearg is None or parsearg(arg):
            print("Don't know what to do with argument %r."%(arg,))
            exit(1)

    _argsparsed = True

    return _verbose, _testmode

# E.g.
#
#   command  = None
#   filename = "default_file"
#
#   def parsearg(arg):
#       global command
#       if command is None:
#           command = arg
#       else:
#           return True
#
#   def parsearg(arg):
#       global filename
#       if arg == '-f':
#           filename = nextarg()
#       else:
#           return True
#
#   verbose, testmode = argparse(parsearg, parseflag, "Use: zooperdoo [-v (verbose)] [-t (test mode)] [-f <file> (use file)] [--] <command>")


#===================================================================
# misc funcs.

def verbose_testmode():
    assert _argsparsed  # Meaningless result before args are parsed.
    return _verbose, _testmode

def echo(s, end='\n'):
    print("## %s"%(str(s).replace('\n','\n## ')), flush=True, end=end)

def bail(s, exitcode=1, test_ignore=False):
    echo(s)
    if _testmode and test_ignore:
        echo("  [TESTMODE: Continuing anyway...]")
        return
    exit(exitcode)

def prompt(s):
    return input("## " + s)
    # echo(s, end='')

def user_wants_to(s):
    if _testmode:
        echo("Ready to %s? [y/N]  [TESTMODE: Continuing...]"%(s,))
        return True
    answer = prompt("Ready to %s? [y/N] "%(s,))
    return answer == 'y'

def file_exists(path):
    "Returns true if path refers to any existing object (including a dead symlink)"
    try:
        os.lstat(path)
        return True
    except:
        return False

def isblockspecial(path):
    try:
        return stat.S_ISBLK(os.stat(path).st_mode)
    except:
        return False

def isdir(path):
    try:
        return stat.S_ISDIR(os.stat(path).st_mode)
    except:
        return False

def file_body(filename, maxsize=-1):
    "Returns the body of file at filename, up to maxsize bytes."
    with open(filename) as fl:
        return fl.read(maxsize)

def file_body_stripped(filename, expect="something informative", maxsize=256):
    """Returns the contents of a file, stripped.  Reads up to maxsize bytes.
    Intended for reading short ids or state.
    """
    if not file_exists(filename):
        bail("ERROR: %s not found.  Expected to contain %s."%(filename, expect))
    return file_body(filename, maxsize).strip()

def assert_file_contains(filename, body):
    if file_body_stripped(filename, body, len(body)*5) != body:
        bail("ERROR: %s expected to contain %s but doesn't."%(filename, body), test_ignore=True)

def path_startswith(parent, child):
    """Like str.startswith, but for directory hierarchies.
    E.g.
        path_startswith("/foo", "/foo/bar") is True
        path_startswith("/foo", "/foobar")  is False
    """
    return parent and child and (child==parent or (child.startswith(parent) and (parent == '/' or child[len(parent)] == '/')))

def reroot(newroot, path):
    "Returns absolute path as if '/' were moved to newroot."
    assert path.startswith('/')
    if path == '/':
        return newroot
    if newroot == '/':
        return path
    return newroot+path

def deroot(oldroot, path):
    "Opposite of reroot.  Returns None if path isn't a child of oldroot."
    if path == oldroot:
        return '/'
    if not path.startswith(oldroot):
        return None
    if oldroot == '/':
        return path
    cut = path[len(oldroot):]
    if cut.startswith('/'):
        return cut
    return None

class RunFail(Exception):
    pass

def run(command, input=None, capture=False, quiet=False, safe=False):
    """Runs a command.

    Input, if provided, is given to the program's stdin.
    If capture is True, returns command's stdout as a string.
    Raises RunFail on failure to exec (couldn't execute, or non-zero exit status).
    If (global) testmode is True, this will not actually run anything unless safe is True.
    """
    if _testmode and not safe:
        assert not capture
        echo(">> [WOULD RUN] %s"%(' '.join(command)))
        return None
    if _verbose or not quiet:
        echo(">> %s"%(' '.join(command)))
    try:
        rv = subprocess.run(command, input=input, encoding='utf8', stdout=capture and subprocess.PIPE or None)
    except Exception as e:
        raise RunFail("Unable to execute: %s"%(e,))
    if rv.returncode:
        raise RunFail("Exit code: %s"%(rv.returncode,))
    if capture:
        return rv.stdout
    return None

def rsync(src, dst, excludes=[]):
    """Rsync from src to dst, copying all attributes, with verbose output.  Does not cross partition boundaries.

    Note src directory is mapped onto dst directory, not into it.  I.e. dst becomes a clone of src rather than containing one.

    Excludes is a list of patterns to --exclude.

    Throws RunFail on error.
    """
    assert not dst.endswith('/')
    if not src.endswith('/'):
        src = src + '/'                             # Map source dir onto (not into) target_dir.
    command  = ['rsync','-axAHXvi','--delete']
    #if _verbose:
    #    command += ["-vi"]
    #else:
    #    command += ["--info=progress2,misc,stats,flist"]   # Trying to show % progress over all instead of individual files, but doesn't work well.
    for exclude in excludes:
        command += ['--exclude', exclude]
    if _testmode:
        command += ['-n']
    command += [src, dst]
    return run(command, safe=True)


#=====================================================================
# Utilities for querying block devices
#

last_devices = None  # Cache of last call to get_devices()

def get_devices(refresh=False):
    """Loads the block device list (see lsblk command).

    Return value is a tree of dictionaries.  Pprint to examine...

    If refresh is True, this will force a re-load even if previously loaded,
        otherwise it will return the cached value when available.
    """
    global last_devices
    if refresh or last_devices is None:
        rv = run(['lsblk','-Jb','-o','NAME,FSTYPE,TYPE,MOUNTPOINT,LABEL,UUID,PARTTYPE,PARTLABEL,PARTUUID,RO,SIZE,STATE,LOG-SEC,PHY-SEC'], capture=True, quiet=True, safe=True)
        last_devices = json.loads(rv)['blockdevices']
    return last_devices

def equals(a, b):
    return a == b

def find_devices(field, value, devices=None, match_func=equals):
    """Finds a list of (dictionaries describing) block devices (within devices) with field matching value.  E.g. ('name', 'sda').
    Uses get_devices() if devices is None.
    match_func may be used to compare (ref_value, field_value) instead of equality.
        NOTE match_func will be called with field_value=None if the field is missing.
    """
    devs = []
    if devices is None:
        devices = get_devices()
    for dev in devices:
        if match_func(value, dev.get(field)):
            devs.append(dev)
        children = dev.get('children')
        if children:
            devs += find_devices(field, value, children, match_func)
    return devs

def find_device(field, value, devices=None):
    devs = find_devices(field, value, devices)
    if not devs:
        return None
    if len(devs) == 1:
        return devs[0]
    raise Exception("Found multiple devices with %s=%r"%(field, value))

def find_device_named(name):
    if name.startswith('/dev/'):
        name = name[5:]
    return find_device('name', name)

def device_parent(device, parent=None, children=None):
    """Returns the parent of the given device (must be from the current devices tree), or None if top level.
    Raises exception if none found.
    """
    if children is None:
        assert last_devices
        children = last_devices  # No point in re-loading.
    for child in children:
        if child is device:
            return parent
        if 'children' in child:
            par = device_parent(device, child, child['children'])
            if par is not None:
                return par
    if parent is None:
        raise Exception("Device not from within currently loaded device tree.")

def device_partitions(diskdev):
    return diskdev and diskdev.get('children', []) or []

def device_name(dev):
    "Returns, e.g., /dev/sda1"
    if not dev:
        return None
    devname = dev.get('name')
    if not devname:
        return None
    return '/dev/'+devname

def get_partitions(diskname):
    """Returns the partitions (as a list of descriptive dictionaries) of the named disk.
    May throw various errors.
    """
    return device_partitions(find_device_named(disk))

#=================================================================================
#  Misc tests.
#

if __name__ == "__main__":
    print("Partitions of nvme0n1: %s"%([part['name'] for part in get_partitions('nvme0n1')],))

    print("Parent of sda: %s"%(device_parent(find_device('name', 'sda')),))
    print("Parent of sda1: %s"%(device_parent(find_device('name', 'sda1')),))
    print("Parent of sda2: %s"%(device_parent(find_device('name', 'sda2')),))

